/*
 * This file is a part of morm
 * and is distributed under the 2-Clause BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2018, Tomasz Sowa
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "baseexpression.h"
#include "morm_types.h"
#include "model.h"
#include "utf8/utf8.h"


namespace morm
{


BaseExpression::BaseExpression()
{
	out_stream = nullptr;
	is_first_field = false;
	work_mode = 0;
}

BaseExpression::~BaseExpression()
{
}


void BaseExpression::set_work_mode(int work_mode)
{
	this->work_mode = work_mode;
}


void BaseExpression::generate_from_model(PT::TextStream & stream, Model & model)
{
	this->out_stream = &stream;

	before_generate_from_model();
	model.map_fields();
	after_generate_from_model();
}


void BaseExpression::before_generate_from_model()
{
	is_first_field = true;
}


void BaseExpression::after_generate_from_model()
{
}


bool BaseExpression::can_field_be_generated(bool insertable, bool updatable, bool is_primary_key)
{
	return true;
}


void BaseExpression::field_before()
{
	if( !is_first_field )
	{
		// put a separator between fields
	}
}


void BaseExpression::field_after()
{
	is_first_field = false;
}


void BaseExpression::put_field_name(const wchar_t * field_name)
{
	before_field_name();
	esc(field_name, *out_stream);
	after_field_name();
}


void BaseExpression::before_field_name()
{
}


void BaseExpression::after_field_name()
{
}


void BaseExpression::before_field_value(const std::wstring &)
{
}


void BaseExpression::before_field_value(const std::string &)
{
}


void BaseExpression::after_field_value(const std::wstring &)
{
}


void BaseExpression::after_field_value(const std::string &)
{
}


void BaseExpression::put_name_value_separator()
{
	(*out_stream) << ',';
}


void BaseExpression::esc(char val, PT::TextStream & stream)
{
	stream << val;
}


void BaseExpression::esc(unsigned char val, PT::TextStream & stream)
{
	esc(static_cast<char>(val), stream);
}


void BaseExpression::esc(const std::wstring & val, PT::TextStream & stream)
{
	char utf8_buf[10];

	for(size_t i = 0 ; i < val.size() ; ++i)
	{
		size_t utf8_len = PT::IntToUTF8((int)val[i], utf8_buf, sizeof(utf8_buf));

		for(size_t a = 0 ; a < utf8_len ; ++a)
		{
			esc(utf8_buf[a], stream);
		}
	}
}


void BaseExpression::esc(const wchar_t * val, PT::TextStream & stream)
{
	char utf8_buf[10];

	for(size_t i = 0 ; val[i] != 0 ; ++i)
	{
		size_t utf8_len = PT::IntToUTF8((int)val[i], utf8_buf, sizeof(utf8_buf));

		for(size_t a = 0 ; a < utf8_len ; ++a)
		{
			esc(utf8_buf[a], stream);
		}
	}
}


void BaseExpression::esc(const std::string & val, PT::TextStream & stream)
{
	for(size_t i = 0 ; i < val.size() ; ++i)
	{
		esc(val[i], stream);
	}
}


void BaseExpression::esc(const char * val, PT::TextStream & stream)
{
	for(size_t i = 0 ; val[i] != 0 ; ++i)
	{
		esc(val[i], stream);
	}
}


void BaseExpression::esc(bool val, PT::TextStream & stream)
{
	stream << val;
}


void BaseExpression::esc(short val, PT::TextStream & stream)
{
	stream << val;
}


void BaseExpression::esc(unsigned short val, PT::TextStream & stream)
{
	stream << val;
}


void BaseExpression::esc(int val, PT::TextStream & stream)
{
	stream << val;
}


void BaseExpression::esc(unsigned int val, PT::TextStream & stream)
{
	stream << val;
}


void BaseExpression::esc(long val, PT::TextStream & stream)
{
	stream << val;
}


void BaseExpression::esc(unsigned long val, PT::TextStream & stream)
{
	stream << val;
}


void BaseExpression::esc(long long val, PT::TextStream & stream)
{
	// not implemented in PT::TextStream yet
	//stream << val;
}


void BaseExpression::esc(unsigned long long val, PT::TextStream & stream)
{
	//stream << val;
}


void BaseExpression::esc(float val, PT::TextStream & stream)
{
	stream << val;
}


void BaseExpression::esc(double val, PT::TextStream & stream)
{
	stream << val;
}


void BaseExpression::esc(long double val, PT::TextStream & stream)
{
	//stream << val;
}


void BaseExpression::esc(void* val, PT::TextStream & stream)
{
	stream << val;
}




}

